from utils import *
from analysis import *
from sqlalchemy import *

from consts import *
from sqlManager import *
import globals
import config

class TableDataProvider:
    """
    Parent class for all table data providers
    """
    def __init__(self, orderby = 0, asc = True):
        self.orderby = orderby
        self.asc = asc
        self.bindParams = None

    """
    Template method, childs have to overload methods that they need
    """
    def _load(self):
        self._buildQuery()
        self._validate()
        self._execute()
        self._analysis()

    def _analysis(self):
        return None

    def _execute(self):
        sqlResult = executeQuery(self.query, self.bindParams)
        self.data = self._sqlResultToData(sqlResult)
        self.qtime = sqlResult.time
        sqlResult.close()

    def _validate(self):
        if (self.orderby < 0 or self.orderby >= len(self.cols)):
            globals.logger.log("Bad orderby col number provided for " + str(self))
            self.orderby = 0

    def _sqlResultToData(self, sqlResult):
        data = []
        for line in sqlResult: data.append(list(line))
        return data

class PagedTableDataProvider(TableDataProvider):

    def __init__(self, orderby, asc, page = None, pageSize = config.PAGE_SIZE, dataCount = None):
        TableDataProvider.__init__(self, orderby, asc)
        self.page = page
        self.pageSize = pageSize
        self.dataCount = dataCount

    def _buildQuery(self):
        self._buildMainQuery()
        if self.dataCount == None:
            self._buildCountQuery()

    def _validate(self):
        TableDataProvider._validate(self)
        if self.page != None and self.page < 0:
            globals.logger.log("Bad page number provided for " + str(self))
            self.page = None

    def _execute(self):
        if self.page != None:
            if type(self.mainQuery) == str:
                self.mainQuery = """SELECT * FROM (""" + \
                    """ SELECT a.*, rownum rnum FROM (""" + \
                    str(self.mainQuery) + """) a WHERE rownum <= """ + str(int(self.page * self.pageSize + self.pageSize)) + \
                    """) WHERE rnum > """ + str(int(self.page * self.pageSize))
            else:
                self.mainQuery = self.mainQuery. \
                    apply_labels(). \
                    offset(self.page * self.pageSize). \
                    limit(self.pageSize)
        # globals.logger.log(self.mainQuery)
        sqlResult = executeQuery(self.mainQuery, self.bindParams)
        self.data = self._sqlResultToData(sqlResult)
        self.qtime = sqlResult.time
        sqlResult.close()
        if self.dataCount == None:
            sqlResult = executeQuery(self.countQuery)
            self.dataCount = int(sqlResult.fetchone()[0])
            self.qtime += sqlResult.time
            sqlResult.close()

class StatsDataProvider:

    def __init__(self, title, tableName):
        self.title = title
        self.tableName = tableName
        self.qtime = 0;
        self.data = []
        self.data.append([])
        self.bindParams = None

    def _buildCols(self):
        self.cols = _getStatsCols(self.tableName)

    def _buildQuery(self, status):
        jobTable = Table(self.tableName, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        self.query = select([func.count(jobTable.c.job_status)], jobTable.c.job_status == status)

    def load(self):
        self._buildCols()
        for i in range (JOB_NEW, JOB_ABANDONED + 1):
            self._buildQuery(i)
            sqlResult = executeQuery(self.query, self.bindParams)
            self.data[0].append(sqlResult.fetchone()[0])
            self.qtime += sqlResult.time
            sqlResult.close()

class RepackStatsDataProvider(StatsDataProvider):
    def __init__(self, runid, title = "Repack stats"):
        StatsDataProvider.__init__(self, title, TABLES.REPACK_JOBS)
        self.runid = runid
        StatsDataProvider.load(self)

    def _buildCols(self):
        self.cols = _getStatsCols(self.tableName, self.runid)

    def _buildQuery(self, status):
#        jobTable = Table(self.tableName, globals.meta, autoload = True, schema = config.DB_SCHEMA)
#        assocTable = Table(TABLES.JOB_DATASET_STREAMER_ASSOC, globals.meta, autoload = True, schema = config.DB_SCHEMA)
#        streamerTable = Table(TABLES.STREAMER, globals.meta, autoload = True, schema = config.DB_SCHEMA)
#        self.query = select(
#            [func.count(distinct(jobTable.c.job_id))],
#            and_(
#                jobTable.c.job_status == status,
#                streamerTable.c.run_id == self.runid
#            ),
#            from_obj = [
#                jobTable.outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
#                    outerjoin(streamerTable, assocTable.c.streamer_id == streamerTable.c.streamer_id)
#            ]
#        )
        schema = ""
        if config.DB_SCHEMA != "" and config.DB_SCHEMA != None:
            schema = config.DB_SCHEMA + "."
        jobTable = schema + self.tableName
        assocTable = schema + TABLES.JOB_DATASET_STREAMER_ASSOC
        streamerTable = schema + TABLES.STREAMER
        repackedTable = schema + TABLES.REPACKED

        self.query = """
            SELECT COUNT(DISTINCT(j.job_id)) FROM
            """ + jobTable + """ j
            LEFT OUTER JOIN """ + assocTable + """ a ON a.job_id = j.job_id
            LEFT OUTER JOIN """ + streamerTable + """ s ON s.streamer_id = a.streamer_id
            WHERE j.job_status = :status AND s.run_id = :runid
        """

        self.bindParams = {'runid' : int(self.runid), 'status' : int(status)}

#        globals.logger.log(self.query)

class MergeStatsDataProvider(StatsDataProvider):
    def __init__(self, runid, title = "Merge stats"):
        StatsDataProvider.__init__(self, title, TABLES.MERGE_JOBS)
        self.runid = runid
        StatsDataProvider.load(self)

    def _buildCols(self):
        self.cols = _getStatsCols(self.tableName, self.runid)

    def _buildQuery(self, status):
        jobTable = Table(TABLES.MERGE_JOBS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        assocTable = Table(TABLES.MERGE_JOB_REPACK_ASSOC, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        repackedTable = Table(TABLES.REPACKED, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        self.query = select(
            [func.count(distinct(jobTable.c.job_id))],
            and_(
                jobTable.c.job_status == status,
                repackedTable.c.run_id == self.runid
            ),
            from_obj = [
                jobTable.outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                    outerjoin(repackedTable, assocTable.c.repacked_id == repackedTable.c.repacked_id)
            ]
        )
        # globals.logger.log(self.query)

class RecoStatsDataProvider(StatsDataProvider):
    def __init__(self, runid, title = "Reco stats"):
        StatsDataProvider.__init__(self, title, TABLES.RECO_JOBS)
        self.runid = runid
        StatsDataProvider.load(self)

    def _buildCols(self):
        self.cols = _getStatsCols(self.tableName, self.runid)

    def _buildQuery(self, status):
        jobTable = Table(TABLES.RECO_JOBS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        assocTable = Table(TABLES.RECO_JOB_REPACK_ASSOC, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        repackedTable = Table(TABLES.REPACKED, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        self.query = select(
            [func.count(distinct(jobTable.c.job_id))],
            and_(
                jobTable.c.job_status == status,
                repackedTable.c.run_id == self.runid
            ),
            from_obj = [
                jobTable.outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                    outerjoin(repackedTable, assocTable.c.repacked_id == repackedTable.c.repacked_id)
            ]
        )
        # globals.logger.log(self.query)

class MergeJobsDataProvider(TableDataProvider):

    cols = [
        ["Job ID", None],
        ["Run ID", formRunLink],
        ["Running time", deltaTime],
        ["Repacked", None],
        ["Dataset", None]
    ]

    def __init__(self, orderby = 0, asc = True, status = None, runid = None):
        TableDataProvider.__init__(self, orderby, asc)
        self.status = status
        self.runid = runid
        s = ""
        if JOB_STATUS.has_key(status): s = JOB_STATUS[status]
        TableDataProvider._load(self)
        if (self.status != None):
            self.cols[2] = ["Definition time", formatTime]
            self.title = "Merge " + s + " jobs"
            if (self.runid != None):
                self.title += " for run " + str(self.runid)
        else:
            self.title = str(len(self.data)) + " merge jobs running"
            self.cols[2] = ["Running time", deltaTime]

    def _buildQuery(self):
        jobTable = Table(TABLES.MERGE_JOBS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        statusTable = Table(TABLES.JOB_STATUS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        assocTable = Table(TABLES.MERGE_JOB_REPACK_ASSOC, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        repackTable = Table(TABLES.REPACKED, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        datasetTable = Table(TABLES.PRIMARY_DATASET, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        c = (
            jobTable.c.job_id,
            repackTable.c.run_id,
            jobTable.c.definition_time,
            func.count(assocTable.c.repacked_id),
            datasetTable.c.name,
        )

        if (self.status != None):
            if (self.runid == None):
                cond = jobTable.c.job_status == self.status
            else:
                cond = and_(
                    jobTable.c.job_status == self.status,
                    repackTable.c.run_id == self.runid,
                )
        else:
            cond = or_(
                jobTable.c.job_status == JOB_USED,
                jobTable.c.job_status == JOB_NEW,
            )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.query = select(
            c,
            cond,
            from_obj = [
                jobTable.
                #outerjoin(statusTable).
                outerjoin(assocTable, jobTable.c.job_id == assocTable.c.job_id).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                datasetTable.c.name,
                repackTable.c.run_id,
            ).order_by(o)

#        globals.logger.log(self.query)

class RecoJobsDataProvider(TableDataProvider):

    cols = [
        ["Job ID", None],
        ["Run ID", formRunLink],
        ["Running time", deltaTime],
        ["Repacked", None],
        ["Dataset", None]
    ]

    def __init__(self, orderby = 0, asc = True, status = None, runid = None):
        TableDataProvider.__init__(self, orderby, asc)
        self.status = status
        self.runid = runid
        s = ""
        if JOB_STATUS.has_key(status): s = JOB_STATUS[status]
        TableDataProvider._load(self)
        if (self.status != None):
            self.title = "Reco " + s + " jobs"
            self.cols[2] = ["Definition time", formatTime]
            if (runid != None): self.title + " for run " + str(self.runid)
        else:
            self.title = str(len(self.data)) + " reco jobs running"
            self.cols[2] = ["Running time", deltaTime]

    def _buildQuery(self):
        jobTable = Table(TABLES.RECO_JOBS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        statusTable = Table(TABLES.JOB_STATUS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        assocTable = Table(TABLES.RECO_JOB_REPACK_ASSOC, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        repackTable = Table(TABLES.REPACKED, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        datasetTable = Table(TABLES.PRIMARY_DATASET, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        c = (
             jobTable.c.job_id,
             repackTable.c.run_id,
             jobTable.c.definition_time,
             func.count(assocTable.c.repacked_id),
             datasetTable.c.name,
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        if (self.status != None):
            if (self.runid == None):
                cond = jobTable.c.job_status == self.status
            else:
                cond = and_(
                    jobTable.c.job_status == self.status,
                    repackTable.c.run_id == self.runid,
                )
        else:
            cond = or_(
                jobTable.c.job_status == JOB_USED,
                jobTable.c.job_status == JOB_NEW,
            )

        self.query = select(
            c,
            cond,
            from_obj = [
                jobTable.
                #outerjoin(statusTable).
                outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                datasetTable.c.name,
                repackTable.c.run_id,
            ).order_by(o)

class RepackJobsDataProvider(TableDataProvider):

    cols = [
        ["Job ID", None],
        ["Run ID", formRunLink],
        ["Running time", deltaTime],
        ["Streamers", None],
        ["Stream", None]
        # ["Dataset", None]
    ]

    def __init__(self, orderby = 0, asc = True, status = None, runid = None):
        TableDataProvider.__init__(self, orderby, asc)
        self.status = status
        self.runid = runid
        s = ""
        if JOB_STATUS.has_key(status): s = JOB_STATUS[status]
        TableDataProvider._load(self)
        if (self.status != None):
            self.cols[2] = ["Definition time", formatTime]
            if (runid == None):
                self.title = "Repacker " + s + " jobs"
            else:
                self.title = "Repacker " + s + " jobs for run " + str(runid)
        else:
            self.title = str(len(self.data)) + " repack jobs running"
            self.cols[2] = ["Running time", deltaTime]

    def _buildQuery(self):
        jobTable = Table(TABLES.REPACK_JOBS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        assocTable = Table(TABLES.JOB_DATASET_STREAMER_ASSOC, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        statusTable = Table(TABLES.JOB_STATUS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        datasetTable = Table(TABLES.PRIMARY_DATASET, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        streamerTable = Table(TABLES.STREAMER, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        c = (
             jobTable.c.job_id,
             streamerTable.c.run_id,
             jobTable.c.definition_time,
             func.count(func.distinct(assocTable.c.streamer_id)),
             streamerTable.c.streamname,
             # datasetTable.c.name,
        )

        if (self.status != None):
            if (self.runid == None):
                cond = jobTable.c.job_status == self.status
            else:
                cond = and_(
                    jobTable.c.job_status == self.status,
                    streamerTable.c.run_id == self.runid,
                )
        else:
            cond = or_(
                jobTable.c.job_status == JOB_USED,
                jobTable.c.job_status == JOB_NEW,
            )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.query = select(
            c,
            cond,
            from_obj = [
                jobTable.
                outerjoin(assocTable, jobTable.c.job_id == assocTable.c.job_id).
                # outerjoin(datasetTable, datasetTable.c.dataset_id == assocTable.c.dataset_id).
                outerjoin(streamerTable, streamerTable.c.streamer_id == assocTable.c.streamer_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                # datasetTable.c.name,
                streamerTable.c.run_id,
                streamerTable.c.streamname,
            ).order_by(o)
        # globals.logger.log(self.query)

class RunRepackDataProvider(PagedTableDataProvider):

    cols = (
        ["Id", None],
        ["Filesize", humanFilesize],
        ["Events", None],
        ["Status", None],
        ["Export status", None],
        ["Dataset", None],
        ["Lfn", lambda x: textToImageAlt(x)]
    )

    def __init__(self, runid, orderby = 0, asc = True, page = 0):
        PagedTableDataProvider.__init__(self, orderby, asc, page)
        self.runid = runid
        PagedTableDataProvider._load(self)
        self.title = str(self.dataCount) + " repacked files"

    def _buildCountQuery(self):
        repackedTable = Table(TABLES.REPACKED, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        self.countQuery = select(
            [func.count(repackedTable.c.repacked_id)],
            repackedTable.c.run_id == self.runid
        )

    def _buildMainQuery(self):
        repackedTable = Table(TABLES.REPACKED, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        rStatusTable = Table(TABLES.REPACKED_STATUS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        eStatusTable = Table(TABLES.EXPORT_STATUS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        datasetTable = Table(TABLES.PRIMARY_DATASET, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        c = (
            repackedTable.c.repacked_id,
            repackedTable.c.filesize,
            repackedTable.c.events,
            rStatusTable.c.status,
            eStatusTable.c.status,
            datasetTable.c.name,
            repackedTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            repackedTable.c.run_id == self.runid,
            from_obj = [repackedTable.
                            outerjoin(rStatusTable, rStatusTable.c.id == repackedTable.c.status).
                            outerjoin(eStatusTable, repackedTable.c.export_status == eStatusTable.c.id).
                            outerjoin(datasetTable, datasetTable.c.dataset_id == repackedTable.c.dataset_id)]
            ).order_by(o)

class RunRecoDataProvider(PagedTableDataProvider):

    cols = (
        ["Id", None],
        ["Filesize", humanFilesize],
        ["Events", None],
        ["Status", None],
        ["Export status", None],
        ["Dataset", None],
        ["Lfn", lambda x: textToImageAlt(x)]
    )

    def __init__(self, runid, orderby = 0, asc = True, page = 0):
        PagedTableDataProvider.__init__(self, orderby, asc, page)
        self.runid = runid
        PagedTableDataProvider._load(self)
        self.title = str(self.dataCount) + " reconstructed files"\

    def _buildCountQuery(self):
        recoTable = Table(TABLES.RECONSTRUCTED, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        self.countQuery = select(
            [func.count(recoTable.c.reconstructed_id)],
            recoTable.c.run_id == self.runid
        )

    def _buildMainQuery(self):
        recoTable = Table(TABLES.RECONSTRUCTED, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        rStatusTable = Table(TABLES.RECONSTRUCTED_STATUS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        eStatusTable = Table(TABLES.EXPORT_STATUS, globals.meta, autoload = True, schema = config.DB_SCHEMA)
        datasetTable = Table(TABLES.PRIMARY_DATASET, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        c = (
            recoTable.c.reconstructed_id,
            recoTable.c.filesize,
            recoTable.c.events,
            rStatusTable.c.status,
            eStatusTable.c.status,
            datasetTable.c.name,
            recoTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            recoTable.c.run_id == self.runid,
            from_obj = [
                recoTable.
                outerjoin(rStatusTable, rStatusTable.c.id == recoTable.c.status).
                outerjoin(eStatusTable, recoTable.c.export_status == eStatusTable.c.id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == recoTable.c.dataset_id)
            ]
            ).order_by(o)

class RunStreamerDataProvider(PagedTableDataProvider):

    cols = (
        ["Id", None],
        ["Lumi Id", None],
        ["Filesize", humanFilesize],
        ["Events", None],
        ["Name", None],
        ["E", None],
        ["S", None],
        ["Lfn", lambda x: textToImageAlt(x)]
    )

    def __init__(self, runid, orderby = 0, asc = True, page = 0):
        PagedTableDataProvider.__init__(self, orderby, asc, page)
        self.runid = runid
        PagedTableDataProvider._load(self)
        self.title = str(self.dataCount) + " streamer files"

    def _buildCountQuery(self):
        streamerTable = Table(TABLES.STREAMER, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        self.countQuery = select(
            [func.count(streamerTable.c.streamer_id)],
            streamerTable.c.run_id == self.runid
        )

    def _buildMainQuery(self):
        streamerTable = Table(TABLES.STREAMER, globals.meta, autoload = True, schema = config.DB_SCHEMA)

        c = (
            streamerTable.c.streamer_id,
            streamerTable.c.lumi_id,
            streamerTable.c.filesize,
            streamerTable.c.events,
            streamerTable.c.streamname,
            streamerTable.c.exportable,
            streamerTable.c.splitable,
            streamerTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            streamerTable.c.run_id == self.runid
        ).order_by(o)

class RunsDataProvider(PagedTableDataProvider):
    # Warning: changing cols ordering affects queries and analysis method
    cols = (
        ["Run ID", formRunLink],
        ["Start", timeToClock],
        ["End", timeToClock],
        ["Process", None],
        ["Version", None],
        ["Global tag", None],
        ["Status", None,
            """
                <i style="color:blue;font-size:10px">Active</i> - Run is ongoing and streamer data is still arriving for it<br />
                <i style="color:blue;font-size:10px">CloseOutSegmentInjector</i> - The Tier-0 is trying to make consistent luminosity sections to be repacked (currently we need 4*n streamer files per Luminosity Section)<br />
                <i style="color:blue;font-size:10px">CloseOutScheduler</i> - There is streamer data for the run that has not been assigned to a repack job yet<br />
                <i style="color:blue;font-size:10px">CloseOutRepack</i> - There is streamer data in a run that has not been repacked yet (create repack jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutRepackMerge</i> - There is RAW data in the run that needs to be merged (create repack merge jobs and wait from them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutPromptReco</i> - There is RAW data in the run that needs to be reconstructed (create prompt reco jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">Complete</i> - We have finished processing this run (there could have been failures along the way but we've exhausted all automatic retries)
            """
        ],
        ["Streamers", None],
        ["Repacked", None,
            """
                The status bar and percentage displays how many streamer
                files were repacked<br /><br />The number shows how many
                repacked files were produced
            """
        ],
        ["Reconstructed", None],
    )

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = 10):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize)
        PagedTableDataProvider._load(self)
        self.title = str(self.dataCount) + " runs"

    def _analysis(self):
        for line in self.data:
            runid = line[0]
            numStreamers = line[7]
            numRepacked = line[8]
            numReco = line[9]
            # analyse repacked percentage
            repackedP = formatPercentage(numRepacked, getRunRepackedPercentage(runid, numStreamers))
            # analyse reco percentage
            recoP = numReco
            # recoP = formatPercentage(numReco, getRunRecoPercentage(runid, numRepacked))

            line[8] = repackedP
            line[9] = recoP

    def _buildCountQuery(self):
        runTable = Table('run', globals.meta, autoload = True, schema = config.DB_SCHEMA)

        self.countQuery = select(
            [func.count(runTable.c.run_id)]
        )

    def _buildMainQuery(self):
        runTable = S_TABLES.RUN
        runStatusTable = S_TABLES.RUN_STATUS
        streamerTable = S_TABLES.STREAMER
        repackedTable = S_TABLES.REPACKED
        reconstructedTable = S_TABLES.RECONSTRUCTED

        if self.asc: order = "ASC"
        else: order = "DESC"

        # HACK: query time optimisation
        if self.orderby <= 6:
            self.mainQuery = """
            SELECT sub.run_id, sub.start_time, sub.end_time, sub.process, sub.version, sub.tag, sub.status,
            (SELECT COUNT(s.run_id) FROM """ + streamerTable + """ s WHERE s.run_id = sub.run_id) numStreamers,
            (SELECT COUNT(re.run_id) FROM """ + repackedTable + """ re WHERE re.run_id = sub.run_id) numRepacked,
            (SELECT COUNT(rec.run_id) FROM """ + reconstructedTable + """ rec WHERE rec.run_id = sub.run_id) numReconstructed
            FROM(
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.global_tag tag,
                rs.status status
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order + """
            ) sub"""
        else:
            # slow query
            self.mainQuery = """
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.global_tag,
                rs.status status,
                (SELECT COUNT(s.run_id) FROM """ + streamerTable + """ s WHERE s.run_id = r.run_id) numStreamers,
                (SELECT COUNT(re.run_id) FROM """ + repackedTable + """ re WHERE re.run_id = r.run_id) numRepacked,
                (SELECT COUNT(rec.run_id) FROM """ + reconstructedTable + """ rec WHERE rec.run_id = r.run_id) numReconstructed
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order

#        globals.logger.log(self.query)

class Cols:
    def __init__(self, cols):
        self.cols = cols
        self.colsMap = {}
        for col in self.cols:
            self.colsMap[col.id] = col

    def setCol(self, id, col):
        if type(id) == int:
            self.cols[id] = col
        else:
            for i in range(len(self.cols)):
                if self.cols[i].id == id:
                    self.cols[i] = col
                    break
        raise Exception("Column with id " + id + " not found")

    def getCol(self, id):
        return self.colsMap[id]

class Col:
    def __init__(self, id, title = None, decorator = None, info = None):
        self.id = id
        if self.title == None:
            self.title = self.id
        else:
            self.title = title
        self.decorator = decorator
        self.info = info

##### private #####

def _getStatsCols(tableName, runid = None):
    return (
            ["New jobs", lambda x: formJobTableLink(x, tableName, JOB_NEW, runid)],
            ["Used jobs", lambda x: formJobTableLink(x, tableName, JOB_USED, runid)],
            ["Successful jobs", lambda x: formJobTableLink(x, tableName, JOB_SUCCESS, runid)],
            ["Failed jobs", lambda x: formJobTableLink(x, tableName, JOB_FAILURE, runid)],
            ["Abandoned jobs", lambda x: formJobTableLink(x, tableName, JOB_ABANDONED, runid)],
        )
